﻿<#
.SYNOPSIS
    Configure, Reconfigure or clean backbone configuration on Chelsio T5/T6 network adapters
.DESCRIPTION
    This script can be used for below three operation on a Chelsio T5/T6 network adapters
    Data is currently collected from the following sources:
        - Configure backbone confuguration
		- Clean backbone configuration
		- Reconfigure backbone configuration.

    Backbone configuaration is an implementation of ring toptology on Chelsio T5/T6 Netowrk Adapters

    To run this script, user must be Administrator and should enable remote powershell.
	Remote powershell can be enabled using below command
		Enable-PSRemoting -Force
	Also to enable authentication, you need to add the all the participating computers to the list of 
	trusted hosts for the local computer in WinRM. 
	To do so, type
		winrm s winrm/config/client '@{TrustedHosts="<RemoteComputer>"}'
.PARAMETER Servers
    All the computers participating in backbone should give seprated by comma.
.PARAMETER BaseIpAddress
    Base ip address of format xx.xx.xx.0 so that script can assign ip address for each nodes.
.PARAMETER NetMask
    Network mask for the backbone network
.PARAMETER Username
	Username for the computers in backbone netowork. All computers participating in backbone network should have same username.
.PARAMETER Password
	Password for the computers in backbone netowork. All computers participating in backbone network should have same Password.
.PARAMETER Clean
	If this switch is present, then script will clean backbone configuration is present
.PARAMETER Reconfigure
	If this switch is present, then script will Reconfigure backbone if it already confugred on this computer.
.PARAMETER EnableVerbose
    If this switch is present, all the verbose messages will get printed on the console.
.EXAMPLE
    .\backbone_cfg.ps1 -Servers node1,node2,node3 -BaseIpAddress 10.10.10.0 -Netmask 24 -Username "Adminisrator" -Password "Test12345"
    Runs backbone_cfg and configure backbone on node1,node2,node3 using base ip address 10.10.10.0 and netmask 24
.EXAMPLE
    .\backbone_cfg.ps1 -Servers node1,node2,node3 -BaseIpAddress 10.10.10.0 -Netmask 24 -Username "Adminisrator" -Password "Test12345" -Clean
    Runs backbone_cfg and clean backbone on node1,node2,node3 using base ip address 10.10.10.0 and netmask 24
.EXAMPLE
    .\backbone_cfg.ps1 -Servers node1,node2,node3 -BaseIpAddress 10.10.10.0 -Netmask 24 -Username "Adminisrator" -Password "Test12345" -Reconfigure
    Runs backbone_cfg and reconfigure backbone on node1,node2,node3 using base ip address 10.10.10.0 and netmask 24
.NOTES

.LINK

#>

[cmdletbinding()]
Param(
    [parameter(Mandatory=$true,HelpMessage="Participating nodes name list seperated by ','")]
    [ValidateNotNullOrEmpty()]
    [String[]] $Servers = $env:COMPUTERNAME,

    [parameter(Mandatory=$false,HelpMessage="Base Ip address of format 102.1.1.0")]
    [ValidateNotNullOrEmpty()]
    [String] $BaseIpAddress = "102.1.1.0",

    [parameter(Mandatory=$false,HelpMessage="Network Ip mask")]
    [ValidateNotNullOrEmpty()]
    [Int]$NetMask = 24,

    [parameter(Mandatory=$false,HelpMessage="Common username for the network")]
    [ValidateNotNullOrEmpty()]
    [String]$Username = "Administrator",

    [parameter(Mandatory=$false,HelpMessage="Common password for the network")]
    [ValidateNotNullOrEmpty()]
    [String]$Password = "Test1234",

    [parameter(Mandatory=$false,HelpMessage="Check for cleaning the Backbone configuration")]
    [Switch] $Clean,

    [parameter(Mandatory=$false,HelpMessage="Check for reconfiguring Backbone configuration")]
    [Switch] $Reconfigure,

    [parameter(Mandatory=$false,HelpMessage="Check for verbose ouput to show")]
    [Switch] $EnableVerbose
)

# Global variables
$Global:__Version = "3.4"
$Global:__Username = $Username
$Global:__Password = $Password
$Global:__OldVerbose = $null
$Global:__MpsTcamTable = @()
$Global:__ServDetails = @()

function Add-Credential()
{
    [CmdletBinding()]
    param(
        [ValidateNotNullOrEmpty()]
        [string]$ComputerName,
        [ValidateNotNullOrEmpty()]
        [string]$User = "Administrator",
        [ValidateNotNullOrEmpty()]
        [string]$Password = "Test1234"
    )

    begin 
    {
        [string]$MstscArguments = ''
        switch ($true) 
        {
            {$Admin} {$MstscArguments += '/admin '}
            {$MultiMon} {$MstscArguments += '/multimon '}
            {$FullScreen} {$MstscArguments += '/f '}
            {$Public} {$MstscArguments += '/public '}
            {$Width} {$MstscArguments += "/w:$Width "}
            {$Height} {$MstscArguments += "/h:$Height "}
        }
    }

    process
    {
        foreach ($Computer in $ComputerName) 
        {
                $ProcessInfo = New-Object System.Diagnostics.ProcessStartInfo
                $Process = New-Object System.Diagnostics.Process

                # Remove the port number for CmdKey otherwise credentials are not entered correctly
                if ($Computer.Contains(':')) {
                    $ComputerCmdkey = ($Computer -split ':')[0]
                } else {
                    $ComputerCmdkey = $Computer
                }

                $ProcessInfo.FileName = "$($env:SystemRoot)\system32\cmdkey.exe"
                $ProcessInfo.Arguments = "/generic:TERMSRV/$ComputerCmdkey /user:$User /pass:$Password"
                $ProcessInfo.WindowStyle = [System.Diagnostics.ProcessWindowStyle]::Hidden
                $Process.StartInfo = $ProcessInfo
                if ($PSCmdlet.ShouldProcess($ComputerCmdkey,'Adding credentials to store')) 
                {
                    [void]$Process.Start()
                }
        }
    }
}

function New-NetDetail()
{
  param ($Computer, $IFaceName,  $MACAddress, $IPAddress)

    $OutputObj  = New-Object -Type PSObject            

    $OutputObj | Add-Member -MemberType NoteProperty -Name ComputerName -Value $Computer.ToUpper()
    $OutputObj | Add-Member -MemberType NoteProperty -Name IFaceName -Value $IFaceName
    $OutputObj | Add-Member -MemberType NoteProperty -Name MACAddress -Value $MACAddress
    $OutputObj | Add-Member -MemberType NoteProperty -Name IPAddress -Value $IPAddress

    return $OutputObj
}

function New-MacDetail()
{
    param ([int]$Index, [string]$Mac, [int]$Pf, [int]$Vf)
    
    $OutputObj  = New-Object -Type PSObject            

    $OutputObj | Add-Member -MemberType NoteProperty -Name Index -Value $Index
    $OutputObj | Add-Member -MemberType NoteProperty -Name MACAddress -Value $Mac
    $OutputObj | Add-Member -MemberType NoteProperty -Name PF -Value $Pf
    $OutputObj | Add-Member -MemberType NoteProperty -Name VF -Value $Vf

    return $OutputObj
}

function Get-NetworkDetails()
{
    [CmdletBinding()]
    param(
        [ValidateNotNullOrEmpty()]
        [string] $ComputerName = $env:computername
    )

    $NAME = $ComputerName.ToUpper()

    try 
    {            
        $Networks = Invoke-Command -ComputerName $ComputerName -ScriptBlock {Get-NetAdapter `
            | where{$_.MacAddress -match "00-07-43"} | Sort-Object -Property MacAddress}
    } 
    catch 
    {            
        Write-Warning "Error occurred while querying network details from $ComputerName."            
        exit(1)
    }

    foreach ($Network in $Networks)
    {
        # Clean if already configured
        [array]$Filters = Invoke-Command -ComputerName $ComputerName -ScriptBlock {cxgbtool.exe nic0 hardware ipfilter show}
        $FilterObject = $Filters[3]
        if(-not [string]::IsNullOrEmpty($FilterObject)) {
            $FilterCount = ($Filters.Count - 3) / 2
            for($i = 0;$i -lt $FilterCount;$i++)
            {
                $null = Invoke-Command -ComputerName $ComputerName -ScriptBlock {cxgbtool.exe nic0 hardware ipfilter remove $using:i}
            }
        }
        # Not save if interface not IPEnabled
        $NetMac = $Network.MacAddress -replace "-",":"
        $WmiObj = Get-WmiObject -Class Win32_NetworkAdapterConfiguration -Filter IPEnabled=FALSE -ComputerName $ComputerName `
        | where{$_.macaddress -eq $NetMac}
        if($WmiObj.InterfaceIndex -eq $Network.ifIndex) {
            continue
        }

        $IFaceName = $Network.Name
        $MACAddress  = $Network.MacAddress
        $Object = New-NetDetail $ComputerName $IFaceName $MACAddress
        $Global:__ServDetails += $Object

        return $Object
    }
}

function Test-Network()
{
    [CmdletBinding()]
    param(
        [ValidateNotNullOrEmpty()]
        [string[]] $Servers = $env:COMPUTERNAME,
        [ValidateNotNullOrEmpty()]
        [string] $Username = "Administrator",
        [ValidateNotNullOrEmpty()]
        [string] $Password = "Test1234"
    )

    [bool]$IsSafeNetwork = $true

    foreach($ComputerName in $Servers)
    {
        $NAME = $ComputerName.ToUpper()

        Write-Verbose "Adding credentials for the Server $ComputerName"
        Add-Credential -ComputerName "$ComputerName" -User "$Username" -Password "$Password"
        if(-not (Test-Connection -ComputerName $ComputerName -Count 1 -ea 0))
        {
            Write-Warning "Machine $NAME is not reachable"
            $IsSafeNetwork = $false
        }

        [String] $ServHostname = Invoke-Command -ComputerName $ComputerName -ScriptBlock{hostname}
        $nomatch = $ServHostname.CompareTo($ComputerName)
        if($nomatch -eq 1)
        {
            $validInputs = 'y','Y','n','N'

            Write-Warning -Message "$NAME's hostname is not matching. This will affect scheduling task for reboot reconfiguration."
            $yes = Read-Host -Prompt "Are you sure you want to continue with this ? (y/n)"
            if($validInputs -notcontains $yes)
            {
                Write-Error "Invalid input" -ErrorAction Stop
            }
            if($yes -eq 'n' -or $yes -eq 'N')
            {
                Write-Verbose -Message "Exiting from the script."
                exit(1)
            }
        }

        $null = Get-NetworkDetails -ComputerName $ComputerName
    }

    if($IsSafeNetwork -ne $true)
    {
        Write-Host "All machines in the list should be reachable" -ForegroundColor Red
        exit(1)
    }
}

function Driver-Reload() 
{
    $chel = Get-pnpdevice | where{$_.Description -match "Chelsio" -and $_.Description -match "Enumerator"}`
     | where{$_.Status -eq "OK"}| select InstanceId

    foreach ($Device in $chel){
        $DeviceId = $Device.InstanceId
        $s = Disable-pnpdevice -InstanceId $DeviceId -Confirm:$false
    }

    foreach ($Device in $chel){
        $DeviceId = $Device.InstanceId
        $s = Enable-PnpDevice -InstanceId $DeviceId -Confirm:$false
    }
}

function Increment-IpAddress()
{
    [CmdletBinding()]
    param(
          [ValidateNotNullOrEmpty()]
          [string]$IpAddress
    )

    $ip = ([System.Net.IPAddress]$IpAddress).GetAddressBytes()
    [Array]::Reverse($ip)
    $ip = ([System.Net.IPAddress]($ip -join '.')).Address
    $ip++
    [Array]::Reverse($ip)
    $ip = $ip -join '.'

    return $ip
}

function Change-RingBackboneReg()
{
    [CmdletBinding()]
    param(
        [ValidateNotNullOrEmpty()]
        [string]$ComputerName,
        [ValidateNotNullOrEmpty()]
        [int]$Value,
        [ValidateNotNullOrEmpty()]
        [switch]$CheckKey = $false
    )

    $NAME = $ComputerName.ToUpper()

    [array]$ChelDevs = Invoke-Command -ComputerName $ComputerName -ScriptBlock {Get-PnpDevice -FriendlyName "*Chelsio*Enumerator*" `
        | where{$_.Status -eq 'OK'}}
    if($ChelDevs.Count -gt 1)
    {
        Write-Error -Message "Multiple Chelsio adapter enabled. Disable Ring-Backbone non-participating adapters." -ErrorAction Stop
        exit(1)
    }

    if($ChelDevs.Count -lt 1)
    {
        Write-Error -Message "No Chelsio adapters found." -ErrorAction Stop
        exit(1)
    }

    $InstanceId = $ChelDevs[0].InstanceId
    $DevKey = Invoke-Command -ComputerName $ComputerName -ScriptBlock {Get-PnpDeviceProperty -InstanceId $using:InstanceId `
         -KeyName DEVPKEY_Device_Driver}
    $DevKeyData = $DevKey.Data
    $PropPath = "HKLM:\SYSTEM\CurrentControlSet\Control\Class\" + $DevKeyData
    $Reg = Invoke-Command -ComputerName $ComputerName -ScriptBlock {Get-ItemProperty -Path $using:PropPath -Name RingBackboneCfg}
    if([string]::IsNullOrEmpty($Reg))
    {
        Write-Error -Message "Couldn't find RingBackboneCfg registry entry" -ErrorAction Stop
        exit(1)
    }
    else
    {

        if($CheckKey)
        {
            if($Reg.RingBackboneCfg -eq 0)
            {
                return $false
            }
            else
            {
                return $true
            }
        }

        Write-Verbose -Message "[$NAME]:Changing Ringbackbone registry value to $Value"

        $null = Invoke-Command -ComputerName $ComputerName -ScriptBlock {Set-ItemProperty -Path $using:PropPath -Name RingBackboneCfg `
            -Value $using:Value}
        Write-Verbose "[$NAME]:Reloading driver"
        $null = Invoke-Command -ComputerName $ComputerName -ScriptBlock ${function:Driver-Reload}
    }
}

function Get-MpsTcamDetails()
{
    [CmdletBinding()]
    param(
        [ValidateNotNullOrEmpty()]
        [string]$ServerName
    )

    [int]$Index = 0
    [string]$Mac = $null
    [int]$Pf = 0
    [int]$Vf = 0
    [string]$Replication = $null
    [bool]$ValidEntry = $false

    $MpsTcam = Invoke-Command -ComputerName $ServerName -ScriptBlock {cxgbtool vbd0 cudbg dbg mpstcam}
    foreach ($Entry in $MpsTcam)
    {
        $ValidEntry = $false
        $MpsEntry = $Entry -split ' '
        for($i = 0;$i -lt $MpsEntry.Length;$i++)
        {
            $Elm = $MpsEntry[$i]
            if($Elm.Contains(":") -and $Elm.Length -gt 1)
            {
                $ValidEntry = $true
                $Mac = $Elm
                $Index = $MpsEntry[$i - 1]
            }

            if($ValidEntry -eq $true)
            {
                if($Elm -eq 'Y')
                {
                    $Pf = $MpsEntry[$i + 6]
                    $Vf = $MpsEntry[$i + 8]
                    $Replication = $MpsEntry[$i + 9]
                    break
                }
            }
        }
        
        if($ValidEntry -eq $true)
        {
            if(($Index -gt 255) -and ([string]::IsNullOrEmpty($Replication)))
            {
                continue
            }
            $Global:__MpsTcamTable += New-MacDetail $Index $Mac $Pf $Vf
        }
    }
}

function Prepare-ScheduledTask
{
    [CmdletBinding()]
    param
    (
        [Parameter(Mandatory=$true)]
        [String] $Argument = $null,
        [Parameter(Mandatory=$true)]
        [String] $Username = $null,
        [Parameter(Mandatory=$true)]
        [String] $Password = $null
    )

    $taskName = "BBScheduledTask"
    $psStartInDir = "C:\Windows\System32\WindowsPowerShell\v1.0"
    $task = Get-ScheduledTask -TaskName $taskName -ErrorAction SilentlyContinue
    if ($task -ne $null)
    {
        Unregister-ScheduledTask -TaskName $taskName -Confirm:$false 
    }

    $action = New-ScheduledTaskAction -Execute 'powershell.exe' -Argument $Argument -WorkingDirectory $psStartInDir
    $trigger = New-ScheduledTaskTrigger -AtStartup -RandomDelay 00:01:00
    $settings = New-ScheduledTaskSettingsSet -Compatibility Win8 -WakeToRun
    $principal = New-ScheduledTaskPrincipal -UserId SYSTEM -LogonType ServiceAccount -RunLevel Highest
    $definition = New-ScheduledTask -Action $action -Principal $principal -Trigger $trigger -Settings $settings -Description "Run $($taskName) at startup"

    Register-ScheduledTask -TaskName $taskName -InputObject $definition -User $Username -Password $Password
}

function Remove-ScheduledTask
{
    $taskName = "BBScheduledTask"
    $task = Get-ScheduledTask -TaskName $taskName -ErrorAction SilentlyContinue
    if ($task -ne $null)
    {
        Unregister-ScheduledTask -TaskName $taskName -Confirm:$false 
    }
}

function Register-BBStartupJob
{
    [CmdletBinding()]
    param
    (
        [Parameter(Mandatory=$true)]
        [String] $ServerName = $env:COMPUTERNAME
    )

    $NAME = $ComputerName.ToUpper()
    $taskName = "BBScheduledTask"

    # Copy script to the destined server
    $null = Invoke-Command -ComputerName $ServerName -ScriptBlock {mkdir C:\bb_temp}
    $null = Copy-Item -Path $PSCommandPath -Destination \\$ServerName\c$\bb_temp\ -Force

    $Argument = '-File "C:\bb_temp\backbone_cfg.ps1" -Server ' + "$ServerName" + ' -Username ' + "$Global:__Username" + ' -Password ' + "$Global:__Password" + ' -Reconfigure'

    $null = Invoke-Command -ComputerName $ServerName -ScriptBlock ${Function:Prepare-ScheduledTask} -ArgumentList $Argument,$Global:__Username,$Global:__Password

    Write-Verbose -Message "[$NAME]:Scheduled task '$taskName' is registered successfully"

}

function UnRegister-BBStartupJob
{
    [CmdletBinding()]
    param
    (
        [Parameter(Mandatory=$true)]
        [String] $ServerName = $env:COMPUTERNAME
    )

    $NAME = $ComputerName.ToUpper()
    $taskName = "BBScheduledTask"

    $null = Invoke-Command -ComputerName $ServerName -ScriptBlock ${Function:Remove-ScheduledTask}
    $null = Invoke-Command -ComputerName $ServerName -ScriptBlock {Remove-Item -Path C:\bb_temp -Force -Recurse -ErrorAction SilentlyContinue}

    Write-Verbose -Message "[$NAME]:Scheduled task '$taskName' is un-registered successfully"
}

function Configure-Backbone()
{
    [CmdletBinding()]
    param(
        [ValidateNotNullOrEmpty()]
        [string]$ComputerName,
        [ValidateNotNullOrEmpty()]
        [ipaddress]$IPAddress,
        [ValidateSet('IPv4', 'IPv6')]
        [string]$AddressFamily = 'IPv4',
        [ValidateNotNullOrEmpty()]
        [int]$PrefixLength = 24
    )

    $NAME = $ComputerName.ToUpper()
    Write-Host "Configuring backbone in $NAME started"

    if(-not $Reconfigure)
    {
        foreach($Entry in $Global:__ServDetails)
        {
            if($Entry.ComputerName -eq $ComputerName)
            {
                $Network = $Entry
                break
            }
        }
        $IFaceName = $Network.IFaceName
        $Network.IPAddress = $IPAddress

        Write-Verbose -Message "[$NAME]:Setting IPAddress:$IPAddress,$AddressFamily,$PrefixLength on interface:$IFaceName"
        $null = Invoke-Command -ComputerName $ComputerName -ScriptBlock {New-NetIPAddress -InterfaceAlias "$using:IFaceName" `
             -IPAddress "$using:IPAddress" -AddressFamily $using:AddressFamily -PrefixLength $using:PrefixLength }
    }

    #SetPass filters for this computer
    Write-Verbose -Message "[$NAME]:Setting filters started"
    $Global:__MpsTcamTable = @()
    Get-MpsTcamDetails -ServerName $ComputerName
    for($i = 1;$i -lt $Global:__MpsTcamTable.Length;$i++)
    {
        $Index = $Global:__MpsTcamTable[$i].Index
        $Pf = $Global:__MpsTcamTable[$i].PF
        $Vf = $Global:__MpsTcamTable[$i].VF
        $Mac = $Global:__MpsTcamTable[$i].MACAddress
        Write-Verbose -Message "[$NAME]:Setting DROP filter for the MAC:$Mac Index:$Index"
        $null = Invoke-Command -ComputerName $ComputerName -ScriptBlock `
        {cxgbtool.exe nic0 hardware ipfilter set -action drop -iport 0 -macidx "$using:Index"}
        if($Index -gt 255)
        {
            Write-Verbose -Message "[$NAME]:Setting DROP filter for the MAC:$Mac Index:$Index in default VI"
            $null = Invoke-Command -ComputerName $ComputerName -ScriptBlock `
            {cxgbtool.exe nic0 hardware ipfilter set -action drop -iport 0 -macidx "$using:Index" -pf $Pf -vf $Vf}
        }
    }
    Write-Verbose -Message "[$NAME]:Setting SWITCH filter"
    Invoke-Command -ComputerName $ComputerName -ScriptBlock `
        {cxgbtool.exe nic0 hardware ipfilter set -action switch -iport 0 -eport 1}

    Write-Host "Configuring backbone in $NAME finished"
}

function Clean-Configuration() 
{
    [CmdletBinding()]
    param(
        [ValidateNotNullOrEmpty()]
        [string]$ComputerName
    )

    $NAME = $ComputerName.ToUpper()

    $Network = Get-NetworkDetails -ComputerName $ComputerName
    $IFaceName = $Network.IFaceName
    $IPAddress = $Network.IPAddress

    Write-Host "Cleaning backbone in $NAME started"

    # Clean the ip address
    Write-Verbose "[$NAME]:Removing IP Address from $IFaceName"
    $s = Invoke-Command -ComputerName $ComputerName -ScriptBlock `
    {Remove-NetIPAddress -InterfaceAlias "$using:IFaceName" -Confirm:$false}
    Write-Verbose "[$NAME]:Reloading driver"
    $s = Invoke-Command -ComputerName $ComputerName -ScriptBlock ${function:Driver-Reload}

    Write-Host "Cleaning backbone in $NAME finished"
}

function Manage-Backbone()
{
    [CmdletBinding()]
    param(
        [parameter(Mandatory=$true,HelpMessage="Participating nodes name list seperated by ','")]
        [ValidateNotNullOrEmpty()]
        [String[]] $Servers = $env:COMPUTERNAME,

        [parameter(Mandatory=$false,HelpMessage="Base Ip address of format 102.1.1.0")]
        [ValidateNotNullOrEmpty()]
        [String] $BaseIpAddress = "102.1.1.0",

        [parameter(Mandatory=$false,HelpMessage="Network Ip mask")]
        [ValidateNotNullOrEmpty()]
        [Int]$NetMask = 24,

        [parameter(Mandatory=$false,HelpMessage="Check for cleaning the Backbone configuration")]
        [Switch] $Clean
    )

    Write-Host "backbone-cfg.ps1 Version $Global:__Version"
    Write-Host "Copyright (c) 2003-2017 Chelsio Communications 2017`n"

    if(-not $Reconfigure)
    {
        if($Clean)
        {
            $yes = Read-Host -Prompt "This will disable ring backbone configuration in Chelsio T5/T6 network adapter. Press 'Y' to proceed"
        }
        else
        {
            $yes = Read-Host -Prompt "This will enable ring backbone configuration in Chelsio T5/T6 network adapter. Press 'Y' to proceed"
        }
        if($yes -ne 'y' -or $yes -ne 'Y') { exit(1) }
    }

    Test-Network -Servers $Servers -Username $Global:__Username -Password $Global:__Password

    [ipaddress]$IPAddress = $BaseIpAddress
    foreach($ComputerName in $Servers)
    {
        $NAME = $ComputerName.ToUpper()

        if($Reconfigure)
        {
            $Ring = Change-RingBackboneReg -ComputerName $ComputerName -CheckKey
            if($Ring -ne $true)
            {
                Write-Error -Message "[$NAME]:[Reconfigure] - RingBackbone is not enabled in this server"
                continue
            }
        }

        if($Clean)
        {
            Clean-Configuration -ComputerName $ComputerName
            Change-RingBackboneReg -ComputerName $ComputerName -Value 0
        }
        else
        {
            if(-not $Reconfigure)
            {
                Change-RingBackboneReg -ComputerName $ComputerName -Value 1
                Start-Sleep -Seconds 5
            }
            $IPAddress = Increment-IpAddress -IpAddress $IPAddress
            Configure-Backbone -ComputerName $ComputerName -IPAddress $IPAddress -AddressFamily IPv4 -PrefixLength 24
        }
    }

    if(-not $Clean)
    {
        Write-Host "`nRing-Backbone details:"
        $Global:__ServDetails
    }
}

if($EnableVerbose) {
    $Global:__OldVerbose = $VerbosePreference
    $VerbosePreference = "continue"
}

# Check is the user is 'Administrator'
$elevated = ([Security.Principal.WindowsPrincipal] `
 [Security.Principal.WindowsIdentity]::GetCurrent()
).IsInRole([Security.Principal.WindowsBuiltInRole]::Administrator)
if($elevated -ne $true)
{
    Write-Error -Message "User must log in as Administrator to run this script." -ErrorAction Stop
}

if($Clean)
{
    Manage-Backbone -Servers $Servers -Clean

    # UnRegister Scheduled job for reconfiguring backbone at reboot startup
    foreach($ComputerName in $Servers)
    {
        UnRegister-BBStartupJob -ServerName $ComputerName
    }
}
else
{
    Manage-Backbone -Servers $Servers -BaseIpAddress $BaseIpAddress -NetMask $NetMask

    if(-not $Reconfigure) 
    {
        # Schedule job to run if reboot happens after confugring backbone
        foreach($ComputerName in $Servers)
        {
            Register-BBStartupJob -ServerName $ComputerName
        }
    }
}

if($EnableVerbose) {
    $VerbosePreference = $Global:__OldVerbose
}
# SIG # Begin signature block
# MIIcrAYJKoZIhvcNAQcCoIIcnTCCHJkCAQExDzANBglghkgBZQMEAgEFADB5Bgor
# BgEEAYI3AgEEoGswaTA0BgorBgEEAYI3AgEeMCYCAwEAAAQQH8w7YFlLCE63JNLG
# KX7zUQIBAAIBAAIBAAIBAAIBADAxMA0GCWCGSAFlAwQCAQUABCAg9gW+VYk7QjUx
# 7EAiQKz4Nms7T8Waoagkvz8YHyO26aCCF78wggUwMIIEGKADAgECAhAECRgbX9W7
# ZnVTQ7VvlVAIMA0GCSqGSIb3DQEBCwUAMGUxCzAJBgNVBAYTAlVTMRUwEwYDVQQK
# EwxEaWdpQ2VydCBJbmMxGTAXBgNVBAsTEHd3dy5kaWdpY2VydC5jb20xJDAiBgNV
# BAMTG0RpZ2lDZXJ0IEFzc3VyZWQgSUQgUm9vdCBDQTAeFw0xMzEwMjIxMjAwMDBa
# Fw0yODEwMjIxMjAwMDBaMHIxCzAJBgNVBAYTAlVTMRUwEwYDVQQKEwxEaWdpQ2Vy
# dCBJbmMxGTAXBgNVBAsTEHd3dy5kaWdpY2VydC5jb20xMTAvBgNVBAMTKERpZ2lD
# ZXJ0IFNIQTIgQXNzdXJlZCBJRCBDb2RlIFNpZ25pbmcgQ0EwggEiMA0GCSqGSIb3
# DQEBAQUAA4IBDwAwggEKAoIBAQD407Mcfw4Rr2d3B9MLMUkZz9D7RZmxOttE9X/l
# qJ3bMtdx6nadBS63j/qSQ8Cl+YnUNxnXtqrwnIal2CWsDnkoOn7p0WfTxvspJ8fT
# eyOU5JEjlpB3gvmhhCNmElQzUHSxKCa7JGnCwlLyFGeKiUXULaGj6YgsIJWuHEqH
# CN8M9eJNYBi+qsSyrnAxZjNxPqxwoqvOf+l8y5Kh5TsxHM/q8grkV7tKtel05iv+
# bMt+dDk2DZDv5LVOpKnqagqrhPOsZ061xPeM0SAlI+sIZD5SlsHyDxL0xY4PwaLo
# LFH3c7y9hbFig3NBggfkOItqcyDQD2RzPJ6fpjOp/RnfJZPRAgMBAAGjggHNMIIB
# yTASBgNVHRMBAf8ECDAGAQH/AgEAMA4GA1UdDwEB/wQEAwIBhjATBgNVHSUEDDAK
# BggrBgEFBQcDAzB5BggrBgEFBQcBAQRtMGswJAYIKwYBBQUHMAGGGGh0dHA6Ly9v
# Y3NwLmRpZ2ljZXJ0LmNvbTBDBggrBgEFBQcwAoY3aHR0cDovL2NhY2VydHMuZGln
# aWNlcnQuY29tL0RpZ2lDZXJ0QXNzdXJlZElEUm9vdENBLmNydDCBgQYDVR0fBHow
# eDA6oDigNoY0aHR0cDovL2NybDQuZGlnaWNlcnQuY29tL0RpZ2lDZXJ0QXNzdXJl
# ZElEUm9vdENBLmNybDA6oDigNoY0aHR0cDovL2NybDMuZGlnaWNlcnQuY29tL0Rp
# Z2lDZXJ0QXNzdXJlZElEUm9vdENBLmNybDBPBgNVHSAESDBGMDgGCmCGSAGG/WwA
# AgQwKjAoBggrBgEFBQcCARYcaHR0cHM6Ly93d3cuZGlnaWNlcnQuY29tL0NQUzAK
# BghghkgBhv1sAzAdBgNVHQ4EFgQUWsS5eyoKo6XqcQPAYPkt9mV1DlgwHwYDVR0j
# BBgwFoAUReuir/SSy4IxLVGLp6chnfNtyA8wDQYJKoZIhvcNAQELBQADggEBAD7s
# DVoks/Mi0RXILHwlKXaoHV0cLToaxO8wYdd+C2D9wz0PxK+L/e8q3yBVN7Dh9tGS
# dQ9RtG6ljlriXiSBThCk7j9xjmMOE0ut119EefM2FAaK95xGTlz/kLEbBw6RFfu6
# r7VRwo0kriTGxycqoSkoGjpxKAI8LpGjwCUR4pwUR6F6aGivm6dcIFzZcbEMj7uo
# +MUSaJ/PQMtARKUT8OZkDCUIQjKyNookAv4vcn4c10lFluhZHen6dGRrsutmQ9qz
# sIzV6Q3d9gEgzpkxYz0IGhizgZtPxpMQBvwHgfqL2vmCSfdibqFT+hKUGIUukpHq
# aGxEMrJmoecYpJpkUe8wggVIMIIEMKADAgECAhALdKznYTakkxmpYQTW0eQyMA0G
# CSqGSIb3DQEBCwUAMHIxCzAJBgNVBAYTAlVTMRUwEwYDVQQKEwxEaWdpQ2VydCBJ
# bmMxGTAXBgNVBAsTEHd3dy5kaWdpY2VydC5jb20xMTAvBgNVBAMTKERpZ2lDZXJ0
# IFNIQTIgQXNzdXJlZCBJRCBDb2RlIFNpZ25pbmcgQ0EwHhcNMTcwNTMwMDAwMDAw
# WhcNMjAwNjAzMTIwMDAwWjCBhDELMAkGA1UEBhMCVVMxEzARBgNVBAgTCkNhbGlm
# b3JuaWExEjAQBgNVBAcTCVN1bm55dmFsZTElMCMGA1UEChMcQ2hlbHNpbyBDb21t
# dW5pY2F0aW9ucywgSW5jLjElMCMGA1UEAxMcQ2hlbHNpbyBDb21tdW5pY2F0aW9u
# cywgSW5jLjCCASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBALOOGpgqi3hI
# QGz7EpFXaQ9PXoaAdIiy1V7VDQkL0YyaDcE0PowyEk93E1pcQbYyVFHPjaF5mOYI
# suNazdtmzhk+4WIdtVHcahh0qKd6eB2oXFKrMQGFz7NqSzakC/fXeK5q8yDki1ET
# ZzNcfSM5ddg40JyLdqLF4bmfcYSCdfv9jUdndhCDGLsrLogMk5gLL+QR15mikE8E
# O/KEyZxxfwkShP3jtpHMoWp0qA7KxZWDawAuW0bGYRLSDYxwYnspFnwEEMt1N/xJ
# 5IJRMEtgLlNEUDBPZ5zPrhWw9IvzR2qdGdBHkTh8HPSbWb83X4DBKdiOIwOOP82x
# K84FG9cB6scCAwEAAaOCAcUwggHBMB8GA1UdIwQYMBaAFFrEuXsqCqOl6nEDwGD5
# LfZldQ5YMB0GA1UdDgQWBBTR3mPAVPQWKq8iymHWZfw2FEaQHTAOBgNVHQ8BAf8E
# BAMCB4AwEwYDVR0lBAwwCgYIKwYBBQUHAwMwdwYDVR0fBHAwbjA1oDOgMYYvaHR0
# cDovL2NybDMuZGlnaWNlcnQuY29tL3NoYTItYXNzdXJlZC1jcy1nMS5jcmwwNaAz
# oDGGL2h0dHA6Ly9jcmw0LmRpZ2ljZXJ0LmNvbS9zaGEyLWFzc3VyZWQtY3MtZzEu
# Y3JsMEwGA1UdIARFMEMwNwYJYIZIAYb9bAMBMCowKAYIKwYBBQUHAgEWHGh0dHBz
# Oi8vd3d3LmRpZ2ljZXJ0LmNvbS9DUFMwCAYGZ4EMAQQBMIGEBggrBgEFBQcBAQR4
# MHYwJAYIKwYBBQUHMAGGGGh0dHA6Ly9vY3NwLmRpZ2ljZXJ0LmNvbTBOBggrBgEF
# BQcwAoZCaHR0cDovL2NhY2VydHMuZGlnaWNlcnQuY29tL0RpZ2lDZXJ0U0hBMkFz
# c3VyZWRJRENvZGVTaWduaW5nQ0EuY3J0MAwGA1UdEwEB/wQCMAAwDQYJKoZIhvcN
# AQELBQADggEBACyYaWjUeNed5pYA5HHtzbTzMiER2V2o5EGGumue10Df/31Kltrn
# J7X06Hg+z3ae+E1tVrghGFsA+mQSmAuQ0hzANdgTp2bb2dF/3vT9kKQpue81UrFz
# RiSjrm1PCd8/dkrlfs4zarivcFsL0l1214j+7JXrvttqtuX9Q6wXDF58hdwuGt67
# pH8jTaqoBTqQjDSwjPOLaXF4llQ4EkWhjczz0AhZKxF4FBHxXfC6ilkHopEj4HE5
# xPm0vUkOmPDp/9UiZrVffMAef7xnRzRB660RaIngku2MwXM/qTX14JyDytoq5xzy
# q2FOEn3uQwV6HnoJrempUa91muQZX6xjotkwggZqMIIFUqADAgECAhADAZoCOv9Y
# sWvW1ermF/BmMA0GCSqGSIb3DQEBBQUAMGIxCzAJBgNVBAYTAlVTMRUwEwYDVQQK
# EwxEaWdpQ2VydCBJbmMxGTAXBgNVBAsTEHd3dy5kaWdpY2VydC5jb20xITAfBgNV
# BAMTGERpZ2lDZXJ0IEFzc3VyZWQgSUQgQ0EtMTAeFw0xNDEwMjIwMDAwMDBaFw0y
# NDEwMjIwMDAwMDBaMEcxCzAJBgNVBAYTAlVTMREwDwYDVQQKEwhEaWdpQ2VydDEl
# MCMGA1UEAxMcRGlnaUNlcnQgVGltZXN0YW1wIFJlc3BvbmRlcjCCASIwDQYJKoZI
# hvcNAQEBBQADggEPADCCAQoCggEBAKNkXfx8s+CCNeDg9sYq5kl1O8xu4FOpnx9k
# WeZ8a39rjJ1V+JLjntVaY1sCSVDZg85vZu7dy4XpX6X51Id0iEQ7Gcnl9ZGfxhQ5
# rCTqqEsskYnMXij0ZLZQt/USs3OWCmejvmGfrvP9Enh1DqZbFP1FI46GRFV9GIYF
# jFWHeUhG98oOjafeTl/iqLYtWQJhiGFyGGi5uHzu5uc0LzF3gTAfuzYBje8n4/ea
# 8EwxZI3j6/oZh6h+z+yMDDZbesF6uHjHyQYuRhDIjegEYNu8c3T6Ttj+qkDxss5w
# RoPp2kChWTrZFQlXmVYwk/PJYczQCMxr7GJCkawCwO+k8IkRj3cCAwEAAaOCAzUw
# ggMxMA4GA1UdDwEB/wQEAwIHgDAMBgNVHRMBAf8EAjAAMBYGA1UdJQEB/wQMMAoG
# CCsGAQUFBwMIMIIBvwYDVR0gBIIBtjCCAbIwggGhBglghkgBhv1sBwEwggGSMCgG
# CCsGAQUFBwIBFhxodHRwczovL3d3dy5kaWdpY2VydC5jb20vQ1BTMIIBZAYIKwYB
# BQUHAgIwggFWHoIBUgBBAG4AeQAgAHUAcwBlACAAbwBmACAAdABoAGkAcwAgAEMA
# ZQByAHQAaQBmAGkAYwBhAHQAZQAgAGMAbwBuAHMAdABpAHQAdQB0AGUAcwAgAGEA
# YwBjAGUAcAB0AGEAbgBjAGUAIABvAGYAIAB0AGgAZQAgAEQAaQBnAGkAQwBlAHIA
# dAAgAEMAUAAvAEMAUABTACAAYQBuAGQAIAB0AGgAZQAgAFIAZQBsAHkAaQBuAGcA
# IABQAGEAcgB0AHkAIABBAGcAcgBlAGUAbQBlAG4AdAAgAHcAaABpAGMAaAAgAGwA
# aQBtAGkAdAAgAGwAaQBhAGIAaQBsAGkAdAB5ACAAYQBuAGQAIABhAHIAZQAgAGkA
# bgBjAG8AcgBwAG8AcgBhAHQAZQBkACAAaABlAHIAZQBpAG4AIABiAHkAIAByAGUA
# ZgBlAHIAZQBuAGMAZQAuMAsGCWCGSAGG/WwDFTAfBgNVHSMEGDAWgBQVABIrE5iy
# mQftHt+ivlcNK2cCzTAdBgNVHQ4EFgQUYVpNJLZJMp1KKnkag0v0HonByn0wfQYD
# VR0fBHYwdDA4oDagNIYyaHR0cDovL2NybDMuZGlnaWNlcnQuY29tL0RpZ2lDZXJ0
# QXNzdXJlZElEQ0EtMS5jcmwwOKA2oDSGMmh0dHA6Ly9jcmw0LmRpZ2ljZXJ0LmNv
# bS9EaWdpQ2VydEFzc3VyZWRJRENBLTEuY3JsMHcGCCsGAQUFBwEBBGswaTAkBggr
# BgEFBQcwAYYYaHR0cDovL29jc3AuZGlnaWNlcnQuY29tMEEGCCsGAQUFBzAChjVo
# dHRwOi8vY2FjZXJ0cy5kaWdpY2VydC5jb20vRGlnaUNlcnRBc3N1cmVkSURDQS0x
# LmNydDANBgkqhkiG9w0BAQUFAAOCAQEAnSV+GzNNsiaBXJuGziMgD4CH5Yj//7HU
# aiwx7ToXGXEXzakbvFoWOQCd42yE5FpA+94GAYw3+puxnSR+/iCkV61bt5qwYCbq
# aVchXTQvH3Gwg5QZBWs1kBCge5fH9j/n4hFBpr1i2fAnPTgdKG86Ugnw7HBi02JL
# sOBzppLA044x2C/jbRcTBu7kA7YUq/OPQ6dxnSHdFMoVXZJB2vkPgdGZdA0mxA5/
# G7X1oPHGdwYoFenYk+VVFvC7Cqsc21xIJ2bIo4sKHOWV2q7ELlmgYd3a822iYemK
# C23sEhi991VUQAOSK2vCUcIKSK+w1G7g9BQKOhvjjz3Kr2qNe9zYRDCCBs0wggW1
# oAMCAQICEAb9+QOWA63qAArrPye7uhswDQYJKoZIhvcNAQEFBQAwZTELMAkGA1UE
# BhMCVVMxFTATBgNVBAoTDERpZ2lDZXJ0IEluYzEZMBcGA1UECxMQd3d3LmRpZ2lj
# ZXJ0LmNvbTEkMCIGA1UEAxMbRGlnaUNlcnQgQXNzdXJlZCBJRCBSb290IENBMB4X
# DTA2MTExMDAwMDAwMFoXDTIxMTExMDAwMDAwMFowYjELMAkGA1UEBhMCVVMxFTAT
# BgNVBAoTDERpZ2lDZXJ0IEluYzEZMBcGA1UECxMQd3d3LmRpZ2ljZXJ0LmNvbTEh
# MB8GA1UEAxMYRGlnaUNlcnQgQXNzdXJlZCBJRCBDQS0xMIIBIjANBgkqhkiG9w0B
# AQEFAAOCAQ8AMIIBCgKCAQEA6IItmfnKwkKVpYBzQHDSnlZUXKnE0kEGj8kz/E1F
# kVyBn+0snPgWWd+etSQVwpi5tHdJ3InECtqvy15r7a2wcTHrzzpADEZNk+yLejYI
# A6sMNP4YSYL+x8cxSIB8HqIPkg5QycaH6zY/2DDD/6b3+6LNb3Mj/qxWBZDwMiEW
# icZwiPkFl32jx0PdAug7Pe2xQaPtP77blUjE7h6z8rwMK5nQxl0SQoHhg26Ccz8m
# SxSQrllmCsSNvtLOBq6thG9IhJtPQLnxTPKvmPv2zkBdXPao8S+v7Iki8msYZbHB
# c63X8djPHgp0XEK4aH631XcKJ1Z8D2KkPzIUYJX9BwSiCQIDAQABo4IDejCCA3Yw
# DgYDVR0PAQH/BAQDAgGGMDsGA1UdJQQ0MDIGCCsGAQUFBwMBBggrBgEFBQcDAgYI
# KwYBBQUHAwMGCCsGAQUFBwMEBggrBgEFBQcDCDCCAdIGA1UdIASCAckwggHFMIIB
# tAYKYIZIAYb9bAABBDCCAaQwOgYIKwYBBQUHAgEWLmh0dHA6Ly93d3cuZGlnaWNl
# cnQuY29tL3NzbC1jcHMtcmVwb3NpdG9yeS5odG0wggFkBggrBgEFBQcCAjCCAVYe
# ggFSAEEAbgB5ACAAdQBzAGUAIABvAGYAIAB0AGgAaQBzACAAQwBlAHIAdABpAGYA
# aQBjAGEAdABlACAAYwBvAG4AcwB0AGkAdAB1AHQAZQBzACAAYQBjAGMAZQBwAHQA
# YQBuAGMAZQAgAG8AZgAgAHQAaABlACAARABpAGcAaQBDAGUAcgB0ACAAQwBQAC8A
# QwBQAFMAIABhAG4AZAAgAHQAaABlACAAUgBlAGwAeQBpAG4AZwAgAFAAYQByAHQA
# eQAgAEEAZwByAGUAZQBtAGUAbgB0ACAAdwBoAGkAYwBoACAAbABpAG0AaQB0ACAA
# bABpAGEAYgBpAGwAaQB0AHkAIABhAG4AZAAgAGEAcgBlACAAaQBuAGMAbwByAHAA
# bwByAGEAdABlAGQAIABoAGUAcgBlAGkAbgAgAGIAeQAgAHIAZQBmAGUAcgBlAG4A
# YwBlAC4wCwYJYIZIAYb9bAMVMBIGA1UdEwEB/wQIMAYBAf8CAQAweQYIKwYBBQUH
# AQEEbTBrMCQGCCsGAQUFBzABhhhodHRwOi8vb2NzcC5kaWdpY2VydC5jb20wQwYI
# KwYBBQUHMAKGN2h0dHA6Ly9jYWNlcnRzLmRpZ2ljZXJ0LmNvbS9EaWdpQ2VydEFz
# c3VyZWRJRFJvb3RDQS5jcnQwgYEGA1UdHwR6MHgwOqA4oDaGNGh0dHA6Ly9jcmwz
# LmRpZ2ljZXJ0LmNvbS9EaWdpQ2VydEFzc3VyZWRJRFJvb3RDQS5jcmwwOqA4oDaG
# NGh0dHA6Ly9jcmw0LmRpZ2ljZXJ0LmNvbS9EaWdpQ2VydEFzc3VyZWRJRFJvb3RD
# QS5jcmwwHQYDVR0OBBYEFBUAEisTmLKZB+0e36K+Vw0rZwLNMB8GA1UdIwQYMBaA
# FEXroq/0ksuCMS1Ri6enIZ3zbcgPMA0GCSqGSIb3DQEBBQUAA4IBAQBGUD7Jtygk
# pzgdtlspr1LPUukxR6tWXHvVDQtBs+/sdR90OPKyXGGinJXDUOSCuSPRujqGcq04
# eKx1XRcXNHJHhZRW0eu7NoR3zCSl8wQZVann4+erYs37iy2QwsDStZS9Xk+xBdIO
# PRqpFFumhjFiqKgz5Js5p8T1zh14dpQlc+Qqq8+cdkvtX8JLFuRLcEwAiR78xXm8
# TBJX/l/hHrwCXaj++wc4Tw3GXZG5D2dFzdaD7eeSDY2xaYxP+1ngIw/Sqq4AfO6c
# Qg7PkdcntxbuD8O9fAqg7iwIVYUiuOsYGk38KiGtSTGDR5V3cdyxG0tLHBCcdxTB
# nU8vWpUIKRAmMYIEQzCCBD8CAQEwgYYwcjELMAkGA1UEBhMCVVMxFTATBgNVBAoT
# DERpZ2lDZXJ0IEluYzEZMBcGA1UECxMQd3d3LmRpZ2ljZXJ0LmNvbTExMC8GA1UE
# AxMoRGlnaUNlcnQgU0hBMiBBc3N1cmVkIElEIENvZGUgU2lnbmluZyBDQQIQC3Ss
# 52E2pJMZqWEE1tHkMjANBglghkgBZQMEAgEFAKB8MBAGCisGAQQBgjcCAQwxAjAA
# MBkGCSqGSIb3DQEJAzEMBgorBgEEAYI3AgEEMBwGCisGAQQBgjcCAQsxDjAMBgor
# BgEEAYI3AgEVMC8GCSqGSIb3DQEJBDEiBCDjCFruw1gcCmRqabyOirx29v3vqJS+
# LAQtmK1lIaV1jjANBgkqhkiG9w0BAQEFAASCAQCUge5w70Q1cEwYkGbtwYcMzOtJ
# k4czduxqrbNXt7QPl0MxP/JT+Sc9y/F57I54OzazaC1zYxhpV1npQV1+Rxk3E90a
# i6az7owZTgOyuFY/zDupzyhss+O/XAHae+pnzJ2gmw6hKmX6/nEOFixlGfprA2+U
# gHoRsPtye3amUr/QTwJWXOhRg7S/1g2tunJ8kTluUi7bTH2+JHvHVttSLPzgJQWN
# fxmusyiaQRm9oB7P7EHmw5kVOL9bhM4/KFoBva/dJgiCTf2ug9Bu34F5dETMpa67
# RjEzJruEvpD0scr81fQ4HD5vrMn8LUD7vHx6BrbNW9/nnV2NnK1GNTDKRSGNoYIC
# DzCCAgsGCSqGSIb3DQEJBjGCAfwwggH4AgEBMHYwYjELMAkGA1UEBhMCVVMxFTAT
# BgNVBAoTDERpZ2lDZXJ0IEluYzEZMBcGA1UECxMQd3d3LmRpZ2ljZXJ0LmNvbTEh
# MB8GA1UEAxMYRGlnaUNlcnQgQXNzdXJlZCBJRCBDQS0xAhADAZoCOv9YsWvW1erm
# F/BmMAkGBSsOAwIaBQCgXTAYBgkqhkiG9w0BCQMxCwYJKoZIhvcNAQcBMBwGCSqG
# SIb3DQEJBTEPFw0xNzA5MjYxNTE4MzdaMCMGCSqGSIb3DQEJBDEWBBSDa+eCNby8
# l6NqxqokeUhtys/s1TANBgkqhkiG9w0BAQEFAASCAQB9+pA1IMzKaMgksyoum2QP
# K8FWBjRRAMmgu9CcNUfRubZzdYvC8d9jO1u4MQtWDjJaWnF7b4h/ZBJ+ijqYTWn7
# CAA3eE+KZfzEErfLOZ0wk5b9jnvDq9VnOdFdpsrHcLJCd3nTw27aCV05KahH0YVI
# hzp/Z/DzwM89hHEf+ybGiFVyl3HPucabIS9x1qE8kCrcCdOXkOP8YpehhV4oh5GN
# VFmZm+Uoe0+bRWdvaJoFUbHt6rZH7U58Ih8yfmga5geH5oOTifB2Of5R7QtEyOdj
# R4rV2TkQQMFA07GXUhqIaUYjCXJDPv7b9RW3Pv5v3BMv67KeoIcHa7lZQbyqojzp
# SIG # End signature block
